@@ -1,4 +1,5 @@
 # Copyright (C) 2022 Pavocracy <pavocracy@pm.me>
+# Signed using RSA key 9A5D2D5AA10873B9ABCD92F1D959AEE8875DEEE6
 # This file is part of leetscraper which is released under GPL-2.0 License.
 
 """Leetscraper, a coding challenge webscraper for leetcode, and other websites!
@@ -10,15 +11,18 @@ filtered down to the problem description, and written to a markdown file.
 This scraper currently works for:
 leetcode.com, projecteuler.net, codechef.com, hackerrank.com, codewars.com
 
+It uses ChromeDriver with Selenium to scrape problems. If Chrome is not installed
+on your machine this scraper will raise an Exception and exit without scraping.
 During class instantiation, kwargs can be accepted to define class behaviour.
 Calling class functions in different orders will also change the behaviour of this scraper.
 It was written with automation in mind. If you wish to use these functions individually,
 See related docstrings for help.
 """
 
+import logging
 from subprocess import run
 from sys import platform
-from time import sleep
+from time import time, sleep
 from json import loads
 from os import getcwd, walk, path, makedirs
 from re import sub
@@ -49,13 +53,13 @@ class Leetscraper:
     """
 
     def __init__(self, **kwargs) -> None:
-        self.supported_website = False
+        supported_website = False
         self.website_name = kwargs.get("website_name", "leetcode.com")
         self.scraped_path = kwargs.get("scraped_path", getcwd())
         self.scrape_limit = kwargs.get("scrape_limit", None)
         auto_scrape = kwargs.get("auto_scrape", True)
         if self.website_name == "leetcode.com":
-            self.supported_website = True
+            supported_website = True
             self.website_options = {
                 "difficulty": {1: "EASY", 2: "MEDIUM", 3: "HARD"},
                 "api_url": "https://leetcode.com/api/problems/all/",
@@ -65,7 +69,7 @@ class Leetscraper:
                 },
             }
         if self.website_name == "projecteuler.net":
-            self.supported_website = True
+            supported_website = True
             self.website_options = {
                 "difficulty": {33: "EASY", 66: "MEDIUM", 100: "HARD"},
                 "api_url": "https://projecteuler.net/recent",
@@ -73,7 +77,7 @@ class Leetscraper:
                 "problem_description": {"id": "content"},
             }
         if self.website_name == "codechef.com":
-            self.supported_website = True
+            supported_website = True
             self.website_options = {
                 "difficulty": {1: "SCHOOL", 2: "EASY", 3: "MEDIUM", 4: "HARD"},
                 "api_url": "https://www.codechef.com/api/list/problems/",
@@ -81,7 +85,7 @@ class Leetscraper:
                 "problem_description": {"class": "problem-statement"},
             }
         if self.website_name == "hackerrank.com":
-            self.supported_website = True
+            supported_website = True
             self.website_options = {
                 "categories": [
                     "algorithms",
@@ -92,10 +96,10 @@ class Leetscraper:
                 ],
                 "api_url": "https://www.hackerrank.com/rest/contests/master/tracks/",
                 "base_url": "https://www.hackerrank.com/challenges/",
-                "problem_description": {"class": "challenge-body-html"},
+                "problem_description": {"class": "problem-statement"},
             }
         if self.website_name == "codewars.com":
-            self.supported_website = True
+            supported_website = True
             self.website_options = {
                 "difficulty": {
                     8: "EASY",
@@ -111,17 +115,26 @@ class Leetscraper:
                 "base_url": "https://www.codewars.com/kata/",
                 "problem_description": {"id": "description"},
             }
-        if not self.supported_website:
-            raise Exception(f"{self.website_name} is not a supported website!")
+        self.logger = self.create_logger()
+        if not supported_website:
+            message = f"{self.website_name} is not a supported website!"
+            self.logger.exception(message)
+            raise Exception(message)
         if not path.isdir(self.scraped_path):
             try:
                 makedirs(self.scraped_path)
             except Exception as error:
-                print(f"Could not use path {self.scraped_path}: {error}")
-                print(f"Using {getcwd()} instead!")
+                self.logger.warning(
+                    "Could not use path %s!: %s.\nUsing %s instead!",
+                    self.scraped_path,
+                    error,
+                    getcwd(),
+                )
                 self.scraped_path = getcwd()
         if self.scrape_limit == 0:
-            raise Exception(f"scrape_limit is set to {self.scrape_limit}!")
+            message = "scrape_limit is set to 0!"
+            self.logger.exception(message)
+            raise Exception(message)
         try:
             if platform.startswith("darwin"):
                 check_chrome_version = run(
@@ -147,14 +160,38 @@ class Leetscraper:
             get_version = str(check_chrome_version.stdout)
             self.chrome_version = sub("[^0-9.]+", "", get_version)
         except Exception as error:
-            raise Exception(f"Could not find chrome version! Error: {error}") from error
+            message = f"Could not find chrome version! Error: {error}"
+            self.logger.exception(message)
+            raise Exception(message) from error
         self.errors = 0
         if auto_scrape:
             scraped_problems = self.scraped_problems()
             needed_problems = self.needed_problems(scraped_problems)
             self.scrape_problems(needed_problems)
+            logging.shutdown()
 
-    def create_webdriver(self) -> webdriver:  # type: ignore[valid-type]
+    def create_logger(self) -> logging.Logger:
+        """Creates the logger. All messages to leetscraper.log, INFO and above to console."""
+        logger = logging.getLogger("leetscraper")
+        logger.setLevel(logging.DEBUG)
+        formatting = logging.Formatter(
+            "%(asctime)s [%(levelname)s]: %(message)s", datefmt="%d/%m/%Y %I:%M:%S %p"
+        )
+        if not logger.hasHandlers():
+            file_handler = logging.FileHandler(
+                f"{path.dirname(__file__)}/leetscraper.log", "a"
+            )
+            file_handler.setLevel(logging.DEBUG)
+            file_handler.setFormatter(formatting)
+            stream_handler = logging.StreamHandler()
+            stream_handler.setLevel(logging.INFO)
+            stream_handler.setFormatter(formatting)
+            logger.addHandler(file_handler)
+            logger.addHandler(stream_handler)
+            logger.debug("Creating log file %s/leetscraper.log", path.dirname(__file__))
+        return logger
+
+    def create_webdriver(self) -> webdriver.chrome.webdriver.WebDriver:
         """Instantiates the webdriver with pre-defined options."""
         options = Options()
         options.headless = True
@@ -166,20 +203,22 @@ class Leetscraper:
         service = Service(
             ChromeDriverManager(log_level=0, print_first_line=False).install()
         )
-        driver = webdriver.Chrome(service=service, options=options)  # type: ignore[operator, call-arg]
+        driver = webdriver.Chrome(service=service, options=options)  # type: ignore[call-arg]
         driver.implicitly_wait(0)
+        self.logger.debug("Created %s webdriver for %s", driver.name, self.website_name)
         return driver
 
     def webdriver_quit(self, driver) -> None:
         """Closes the webdriver."""
-        print(f"Closing {self.website_name} driver")
+        self.logger.debug("Closing %s driver", self.website_name)
         driver.quit()
 
     def scraped_problems(self) -> List[str]:
         """Returns a list of all website problems already scraped in the scraped_path."""
-        print(
-            f"Checking {self.scraped_path} for exsisting {self.website_name} problems"
+        self.logger.info(
+            "Checking %s for existing %s problems", self.scraped_path, self.website_name
         )
+        start = time()
         scraped_problems = []
         for (dirpath, dirnames, filenames) in walk(
             f"{self.scraped_path}/PROBLEMS/{self.website_name}"
@@ -196,11 +235,19 @@ class Leetscraper:
                         scraped_problems.append(file.split(".")[0])
                     if self.website_name == "codewars.com":
                         scraped_problems.append(file.split(".")[0])
+        stop = time()
+        self.logger.debug(
+            "checking for %s scraped_problems in %s took %s seconds",
+            self.website_name,
+            self.scraped_path,
+            int(stop - start),
+        )
         return scraped_problems
 
     def needed_problems(self, scraped_problems: list) -> List[List[str]]:
         """Returns a list of website problems missing from the scraped_path."""
-        print(f"Getting the list of {self.website_name} problems to scrape")
+        self.logger.info("Getting the list of %s problems to scrape", self.website_name)
+        start = time()
         http = PoolManager()
         get_problems = []
         if self.website_name == "leetcode.com":
@@ -222,12 +269,12 @@ class Leetscraper:
         if self.website_name == "projecteuler.net":
             request = http.request("GET", self.website_options["api_url"])
             soup = BeautifulSoup(request.data, "html.parser")
-            data = soup.find("td", {"class": "id_column"}).get_text()  # type: ignore[union-attr]
+            data = soup.find("td", {"class": "id_column"}).get_text()
             for i in range(1, int(data) + 1):
                 if str(i) not in scraped_problems:
-                    get_problems.append([str(i), None])  # type: ignore[list-item]
+                    get_problems.append([str(i), None])
         if self.website_name == "codechef.com":
-            for value in self.website_options["difficulty"].values():  # type: ignore[attr-defined]
+            for value in self.website_options["difficulty"].values():
                 request = http.request(
                     "GET",
                     self.website_options["api_url"] + value.lower() + "?limit=999",
@@ -235,7 +282,7 @@ class Leetscraper:
                 data = loads(request.data.decode("utf-8"))
                 for problem in data["data"]:
                     if problem["code"] not in scraped_problems:
-                        get_problems.append([problem["code"], value])  # type: ignore[list-item]
+                        get_problems.append([problem["code"], value])
         if self.website_name == "hackerrank.com":
             headers = {}
             chrome_version = f"Chrome/{self.chrome_version}"
@@ -262,8 +309,8 @@ class Leetscraper:
                     else:
                         break
         if self.website_name == "codewars.com":
-            print(
-                "***INFO*** codewars can take up to 5 minutes to search all problems!"
+            self.logger.info(
+                "**NOTE** codewars can take up to 5 minutes to search all problems!"
             )
             for i in range(0, 999):
                 request = http.request(
@@ -277,8 +324,14 @@ class Leetscraper:
                             get_problems.append([problem["id"], None])
                 else:
                     break
+        stop = time()
+        self.logger.debug(
+            "getting list of needed_problems for %s took %s seconds",
+            self.website_name,
+            int(stop - start),
+        )
         http.clear()
-        return get_problems  # type: ignore[return-value]
+        return get_problems
 
     def scrape_problems(self, needed_problems: List[List[str]]) -> None:
         """Scrapes needed_problems limited by scrape_limit. (All problems if scrape_limit not set)"""
@@ -286,42 +339,65 @@ class Leetscraper:
             if self.scrape_limit >= len(needed_problems):
                 self.scrape_limit = None
         if needed_problems:
-            print(
-                f"Attempting to scrape {self.scrape_limit if self.scrape_limit else len(needed_problems)} {self.website_name} problems"
+            self.logger.info(
+                "Attempting to scrape %s %s problems",
+                self.scrape_limit if self.scrape_limit else len(needed_problems),
+                self.website_name,
             )
+            start = time()
             driver = self.create_webdriver()
             for problem in tqdm(needed_problems[: self.scrape_limit]):
                 self.create_problem(problem, driver)
             self.webdriver_quit(driver)
+            stop = time()
+            self.logger.debug(
+                "scraping %s %s problems took %s seconds",
+                self.scrape_limit if self.scrape_limit else len(needed_problems),
+                self.website_name,
+                int(stop - start),
+            )
         else:
-            print(f"No {self.website_name} problems to scrape")
-        print(
-            f"Successfully scraped {(self.scrape_limit if self.scrape_limit else len(needed_problems)) - self.errors} {self.website_name} Problems"
-        )
+            self.logger.warning("No %s problems to scrape", self.website_name)
+        if self.errors:
+            self.logger.warning(
+                "Scraped %s problems, but %s problems failed! Check leetscraper.log for failed scrapes.",
+                self.scrape_limit
+                if self.scrape_limit
+                else len(needed_problems) - self.errors,
+                self.errors,
+            )
+        else:
+            self.logger.info(
+                "Successfully scraped %s %s problems",
+                self.scrape_limit if self.scrape_limit else len(needed_problems),
+                self.website_name,
+            )
 
-    def create_problem(self, problem: List[str], driver: webdriver) -> None:  # type: ignore[valid-type]
+    def create_problem(
+        self, problem: List[str], driver: webdriver.chrome.webdriver.WebDriver
+    ) -> None:
         """Gets the html source of a problem, filters down to the problem description, creates a file.\n
         Creates files in scraped_path/website_name/DIFFICULTY/problem.md
         """
         try:
-            driver.get(self.website_options["base_url"] + problem[0])  # type: ignore[operator, attr-defined]
+            driver.get(self.website_options["base_url"] + problem[0])  # type: ignore[operator]
             WebDriverWait(driver, 10).until(
                 EC.invisibility_of_element_located((By.ID, "initial-loading")),
                 "Timeout limit reached",
             )
             sleep(1)
-            html = driver.page_source  # type: ignore[attr-defined]
+            html = driver.page_source
             soup = BeautifulSoup(html, "html.parser")
             if self.website_name == "leetcode.com":
                 problem_description = (
-                    soup.find("div", self.website_options["problem_description"])  # type: ignore[union-attr, arg-type]
+                    soup.find("div", self.website_options["problem_description"])
                     .get_text()
                     .strip()
                 )
                 problem_name = problem[0]
             if self.website_name == "projecteuler.net":
                 problem_description = (
-                    soup.find("div", self.website_options["problem_description"])  # type: ignore[union-attr, arg-type]
+                    soup.find("div", self.website_options["problem_description"])
                     .get_text()
                     .strip()
                 )
@@ -338,18 +414,16 @@ class Leetscraper:
                     )
                 except IndexError:
                     difficulty = 100
-                for key, value in self.website_options["difficulty"].items():  # type: ignore[attr-defined]
+                for key, value in self.website_options["difficulty"].items():
                     if int(difficulty) <= key:
                         problem[1] = value
                         break
                 problem_description = (
-                    soup.find("div", {"class": "problem_content"})  # type: ignore[union-attr]
-                    .get_text()
-                    .strip()
+                    soup.find("div", {"class": "problem_content"}).get_text().strip()
                 )
             if self.website_name == "codechef.com":
                 problem_description = (
-                    soup.find("div", self.website_options["problem_description"])  # type: ignore[union-attr, arg-type]
+                    soup.find("div", self.website_options["problem_description"])
                     .get_text()
                     .split("Author:")[0]
                     .strip()
@@ -365,7 +439,7 @@ class Leetscraper:
                 problem_name = problem[0] + f"-{problem_name}"
             if self.website_name == "hackerrank.com":
                 problem_description = (
-                    soup.find("div", self.website_options["problem_description"])  # type: ignore[union-attr, arg-type]
+                    soup.find("div", self.website_options["problem_description"])
                     .get_text()
                     .strip()
                 )
@@ -375,7 +449,7 @@ class Leetscraper:
                     difficulty = self.website_options["difficulty"][  # type: ignore[index]
                         (
                             int(
-                                soup.find("div", {"class": "inner-small-hex"})  # type: ignore[union-attr]
+                                soup.find("div", {"class": "inner-small-hex"})
                                 .get_text()
                                 .split(" ")[0]
                             )
@@ -384,12 +458,12 @@ class Leetscraper:
                 except Exception:
                     difficulty = "BETA"  # type: ignore[assignment]
                 problem_description = (
-                    soup.find("div", self.website_options["problem_description"])  # type: ignore[union-attr, arg-type]
+                    soup.find("div", self.website_options["problem_description"])
                     .get_text()
                     .strip()
                 )
                 problem_name = problem[0]
-                problem[1] = difficulty  # type: ignore [call-overload]
+                problem[1] = difficulty  # type: ignore[call-overload]
             if not path.isdir(
                 f"{self.scraped_path}/PROBLEMS/{self.website_name}/{problem[1]}/"
             ):
@@ -404,7 +478,10 @@ class Leetscraper:
                 file.writelines(self.website_options["base_url"] + problem[0] + "\n\n")  # type: ignore[operator]
                 file.writelines(problem_description + "\n")
         except Exception as error:
-            print(
-                f'\nError occurred while scraping {self.website_options["base_url"]}{problem[0]}: {error}'
+            self.logger.debug(
+                "\nError occurred while scraping %s%s: %s",
+                self.website_options["base_url"],
+                problem[0],
+                error,
             )
             self.errors += 1
